/*
 * checkvic.c - An add-on for VGA-Planets (PHOST >=3)
 *
 * Copyright (C) 1999-2001 Heiko Schlenker
 *
 * Author: Heiko Schlenker <hschlen@gmx.de>
 * Maintainer: Heiko Schlenker <hschlen@gmx.de>
 * Version: $Revision: 1.3 $
 * Status: quick & dirty
 *
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


/*
 * $Id: checkvic.c,v 1.3 2001/03/06 21:26:31 hschlen Exp $
 */

/*
 * $Log: checkvic.c,v $
 * Revision 1.3  2001/03/06 21:26:31  hschlen
 * *** empty log message ***
 *
 * Revision 1.2  2001/03/06 20:11:32  hschlen
 * added replacement snprintf() function
 *
 * Revision 1.1.1.1  2001/02/01 20:35:25  hschlen
 * initial import into CVS
 *
 */



#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <memory.h>
#include <string.h>
#include <unistd.h>
#include <phostpdk.h>

#ifdef __sun__
extern char *sys_errlist[];
#endif


#define MAXMSGLINE 42   /* A VGAP message is assumed to be a plain-text
                           null-terminated string not exceeding 840 characters
                           (20 lines by 42 characters, with a 0Dh at the end of
                           each line). */


static const char rcsid[] = "$Id: checkvic.c,v 1.3 2001/03/06 21:26:31 hschlen Exp $";
const char CHECKVIC_VERSION[] = "1.3";
const char ReleaseDate[] = "$Date: 2001/03/06 21:26:31 $";
const char Notice[] = "Copyright (C) 1999-2001 Heiko Schlenker.\n\
Checkvic comes with ABSOLUTELY NO WARRANTY.\n\
This is free software. See COPYING for details.";



/* Define an array of players. */

struct player {
    Uns16 id;
    Uns16 planetsum;
    Uns16 consecutiveturns;
    Boolean winner;
};

struct player playerslist[] = {
    {0, 0, 0, False},
    {1, 0, 0, False},
    {2, 0, 0, False},
    {3, 0, 0, False},
    {4, 0, 0, False},
    {5, 0, 0, False},
    {6, 0, 0, False},
    {7, 0, 0, False},
    {8, 0, 0, False},
    {9, 0, 0, False},
    {10, 0, 0, False},
    {11, 0, 0, False}
};

int count = sizeof(playerslist) / sizeof(struct player);



/* These are the comparison functions used for sorting and searching. */

int playerfind_cmp(const void *p1, const void *p2)
{
    struct player *sp1 = (struct player *) p1;
    struct player *sp2 = (struct player *) p2;


    return (int) ((sp1->id) - (sp2->id));
}

int player_cmp(const void *p1, const void *p2)
{
    struct player *sp1 = (struct player *) p1;
    struct player *sp2 = (struct player *) p2;


    return (int) ((sp2->planetsum) - (sp1->planetsum));
}


/* Print information about a player. */

/*
 void
 print_player (const struct player *p)
 {
 Info ("Player %u holds %u planets for %u turns\n", p->id, p->planetsum,
 p->consecutiveturns);
 }
 */

/* Do the lookup into the sorted array. */

struct player *find_player(const Uns16 id)
{
    struct player target, *result;


    result = NULL;
    target.id = id;
    result = bsearch(&target, playerslist, count, sizeof(struct player), playerfind_cmp);

    return (result ? result : NULL);
}



/* Main function.*/

int main(int argc, char *argv[])
{
    Uns16 howmanywinners = 0;
    Uns16 planetsforvictory = 0, turnsforvictory = 0;
    Uns16 lastturnnumber = 0, currentturnnumber = 0;
    Uns16 i = 0, race = 0, planetunownedsum = 0;
    Int16 x = 0;
    Int16 limittoshow = 0;
    char planetunownedmessage[41] = { '\0' };
    char *messages[21] = { NULL };
    char *message = NULL;
    const char *messagePtr[1];
    FILE *statusfile = NULL;
    FILE *errorlogfile = NULL;
    char *errorlog = NULL;
    int c = 0, returncode = 0;
    Boolean reportunownedplanets = False;
    Boolean limitreached = False;
    Boolean phostlibstatus = False;


    /*
     * Parse the command-line parameters.
     */

    /*
     * Command line usage is:
     *  checkvic [options] <PlanetsForVictory> <ConsecutiveTurns>
     *                     [<GameDirectory> [<RootDirectory>]]
     *  options:
     *
     *           -h OR -H   -- get help
     *           -v         -- print version
     *           -u         -- reports number of unowned planets
     *           -e file    -- use 'file' as an error log file
     *           -p N       -- specify number N (range [0,100])
     */

    while ((c = pgetopt(argc, argv, "hHvup:e:")) != EOF) {
	switch (c) {
	case 'h':
	case 'H':
	case '?':
	    fprintf(errorlogfile ? errorlogfile : stdout,
		    "\nCheckvic %s (%s)\n", CHECKVIC_VERSION, ReleaseDate);
	    fprintf(errorlogfile ? errorlogfile : stdout, "%s\n\n", Notice);
	    fprintf(errorlogfile ? errorlogfile : stdout,
"Usage: checkvic [options] <PlanetsForVictory> <ConsecutiveTurns> [<GameDirectory> [<RootDirectory>]]\n\
\n\
options:\n\
         -h OR -H     -- prints this help summary and exits\n\
         -v           -- prints program version and exits\n\
         -u           -- reports number of unowned planets\n\
         -e file      -- use 'file' as an error log file\n\
         -p N         -- if a player owns N percent of the required number\n\
                         of planets report the amount he controls\n\
\n\
NOTE: Use this add-on in the AUXHOST2 phase.\n");
	    exit(EXIT_SUCCESS);

	case 'v':
	    printf("checkvic %s (%s)\n", CHECKVIC_VERSION, ReleaseDate);
	    exit(EXIT_SUCCESS);

	case 'u':
	    reportunownedplanets = True;
	    break;

	case 'p':
	    limittoshow = (Int16) atoi(poptarg);
	    if ((limittoshow < 0) || (limittoshow > 100)) {
		fprintf(errorlogfile ? errorlogfile : stderr,
			"Error: '-p' parameter not in range [0,100]!\n       Type '%s -h' for a usage summary.\n",
			argv[0]);
		returncode = EXIT_FAILURE;
		goto error;
	    }
	    break;

	case 'e':
	    if (NULL == (errorlog = MemAlloc(strlen(poptarg) + 1))) {
		fprintf(errorlogfile ? errorlogfile : stderr,
			"(errorlog = MemAlloc(strlen(poptarg)+40)) failed\n");
		returncode = EXIT_FAILURE;
		goto error;
	    } else {
		strcpy(errorlog, poptarg);
		errorlogfile = fopen(errorlog, "w");
		if (!errorlogfile) {
		    fprintf(errorlogfile ? errorlogfile : stderr,
			    "Cannot open error logfile '%s' (%s)!\n", errorlog, strerror(errno));
		    returncode = EXIT_FAILURE;
		    goto error;
		}
	    }
	    break;

	default:
	    ;
	}
    }

    if (poptind < argc) {
	planetsforvictory = (Uns16) atoi(argv[poptind++]);
	if ((0 == planetsforvictory) || (planetsforvictory > (Uns16) PLANET_NR)) {
	    fprintf(errorlogfile ? errorlogfile : stderr,
		    "Error: 'PlanetsForVictory' parameter not in range [0,%u]!\n       Type '%s -h' for a usage summary.\n",
		    (Uns16) PLANET_NR, argv[0]);
	    returncode = EXIT_FAILURE;
	    goto error;
	}
    } else {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Error: Must specify number of planets needed for victory.\n       Type '%s -h' for a usage summary.\n",
		argv[0]);
	returncode = EXIT_FAILURE;
	goto error;
    }

    if (poptind < argc) {
	turnsforvictory = (Uns16) atoi(argv[poptind++]);
	if (0 == turnsforvictory) {
	    fprintf(errorlogfile ? errorlogfile : stderr,
		    "Error: A player has to control the given number of planets\n       for at least 1 turn!\n       Type '%s -h' for a usage summary.\n",
		    argv[0]);
	    returncode = EXIT_FAILURE;
	    goto error;
	}
    } else {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Error: Must specify number of turns a player has to control\n       a distinct number of planets.\n       Type '%s -h' for a usage summary.\n",
		argv[0]);
	returncode = EXIT_FAILURE;
	goto error;
    }

    if (poptind < argc) {
	gGameDirectory = argv[poptind++];
    }
    if (poptind < argc) {
	gRootDirectory = argv[poptind++];
    }
    if (poptind < argc) {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Error: Too many arguments. First extraneous parameter is '%s'.\n       Type '%s -h' for a usage summary.\n",
		argv[poptind], argv[0]);
	returncode = EXIT_FAILURE;
	goto error;
    }


    /*
     * Initialize the PDK
     */
    InitPHOSTLib();
    phostlibstatus = True;

    if (!ReadGlobalData()) {
	returncode = EXIT_FAILURE;
	goto error;
    }
    if (!ReadHostData()) {
	returncode = EXIT_FAILURE;
	goto error;
    }


    /*
     * Now go through and count items for each player. First initialize
     * all counts to 0.
     */
    memset(messages, 0, sizeof(messages));
    memset(messagePtr, 0, sizeof(messagePtr));
    message = NULL;
    lastturnnumber = 0;
    currentturnnumber = TurnNumber();
    statusfile = NULL;
    limitreached = False;


    /*
     * Open the log file.
     */
    gLogFile = OpenOutputFile("checkvic.log", TEXT_MODE | GAME_DIR_ONLY | NO_MISSING_ERROR);
    if (!gLogFile) {
	fprintf(errorlogfile ? errorlogfile : stderr, "Cannot open checkvic.log!\n");
	returncode = EXIT_FAILURE;
	goto error;
    }


    /*
     * Allocate memory.
     */
    if (NULL == (message = MemAlloc(840))) {
	fprintf(errorlogfile ? errorlogfile : stderr, "(message = MemAlloc(840)) failed\n");
	returncode = EXIT_FAILURE;
	goto error;
    }
    message[0] = '\0';

    for (i = 0; i < 20; i++) {
	if (NULL == (messages[i] = MemAlloc(MAXMSGLINE))) {
	    fprintf(errorlogfile ? errorlogfile : stderr,
		    "(messages[%i] = MemAlloc(MAXMSGLINE)) failed\n", i);
	    MemFree(message);
	    for (x = i - 1; x >= 0; x--) {
		MemFree(messages[x]);
	    }
	    returncode = EXIT_FAILURE;
	    goto error;
	}
	strcpy(messages[i], "\0");
    }


    /*
     * Sort the players list by comparing
     * the id of each player.
     */
    qsort(playerslist, count, sizeof(struct player), playerfind_cmp);


    /*
     * Read the data file.
     */
    statusfile = OpenInputFile("checkvic.hst", GAME_DIR_ONLY | NO_MISSING_ERROR);
    if (statusfile) {
	if(1 != (fread(&lastturnnumber, sizeof(Uns16), 1, statusfile))) {
	    fprintf(errorlogfile ? errorlogfile : stderr, "checkvic.hst: Cannot get last turn number!\n");
	    fclose(statusfile);
	    statusfile = NULL;
	    returncode = EXIT_FAILURE;
	    goto error;
	}
	if (1 != (fread(playerslist, sizeof(playerslist), 1, statusfile))) {
	    fprintf(errorlogfile ? errorlogfile : stderr, "checkvic.hst: Cannot get last checkvic status!\n");
	    fclose(statusfile);
	    statusfile = NULL;
	    returncode = EXIT_FAILURE;
	    goto error;
	}
	fclose(statusfile);
	statusfile = NULL;
    } else {
	fprintf(errorlogfile ? errorlogfile : stderr, "Warning: Cannot open checkvic.hst for reading!\n");
    }


    /*
     * Count planets.
     */
    for (race = 0; race <= RACE_NR; race++) {
	find_player(race)->planetsum = 0;
    }
    for (i = 1; i <= PLANET_NR; i++) {
	if (IsPlanetExist(i)) {
	    find_player(PlanetOwner(i))->planetsum++;
	    if (True == reportunownedplanets) {
		if (NoOwner == PlanetOwner(i)) {
		    planetunownedsum++;
		}
	    }
	}
    }


    /*
     * Generate report.
     */
    Info("Checkvic %s (%s)", CHECKVIC_VERSION, ReleaseDate);
#if !defined (checkvicBSD) && defined (unix)
    Info("   Running: %s%s -p %u %u %u %s %s\n", basename(argv[0]), reportunownedplanets ? " -u" : "", limittoshow,
	 planetsforvictory, turnsforvictory, basename(gGameDirectory),
	 basename(gRootDirectory));
#else
    Info("   Running: %s%s -p %u %u %u %s %s\n", argv[0], reportunownedplanets ? " -u" : "", limittoshow,
	 planetsforvictory, turnsforvictory, gGameDirectory,
	 gRootDirectory);
#endif
    Info("Turn: %u", currentturnnumber);
    Info("Victory conditions:\n    To hold %u planets for %u consecutive turns.\n",
	 planetsforvictory, turnsforvictory);
    sprintf(messages[0], "(-a0000)<<< Checkvictory Report >>>\x0D");
    strcat(message, messages[0]);
    snprintf(messages[1], MAXMSGLINE, "Turn: %u\x0D", currentturnnumber);
    strcat(message, messages[1]);
    sprintf(messages[2], "Victory conditions:\x0D");
    strcat(message, messages[2]);
    snprintf(messages[3], MAXMSGLINE, "  To hold %u planets for\x0D", planetsforvictory);
    strcat(message, messages[3]);
    snprintf(messages[4], MAXMSGLINE, "  %u consecutive turns.\x0D", turnsforvictory);
    strcat(message, messages[4]);
    sprintf(messages[5], "\x0D");
    strcat(message, messages[5]);

    /*
     * Check if a race holds more than "limittoshow" percent
     * of the necessary planets.
     */
    for (race = 0; race <= RACE_NR; race++) {
	if (playerslist[race].id > 0) { /* Race 0 is nothing. */
	    if (((find_player(race)->planetsum * 100) / planetsforvictory) >= limittoshow) {
		limitreached = True;
	    }
	    if (playerslist[race].planetsum < planetsforvictory) {
		playerslist[race].consecutiveturns = 0;
	    }
	}
    }

    /*
     * Sort the players list by comparing the
     * amount of planets each player holds.
     */
    qsort(playerslist, count, sizeof(struct player), player_cmp);

    /*
     * Check if a report should be generated.
     */
    if (True == limitreached) {
	Info("\nPLAYER                    PLANETS   CONSECUTIVE TURNS");
	sprintf(messages[6], "PLAYER                 PLANETS  TURNS\x0D");
	strcat(message, messages[6]);

	for (i = 0; i <= RACE_NR; i++) {

	    /* Remember: race 0 is nothing.
	     * Since the array is sorted now
	     * playerslist[0] maybe does not
	     * respresent race 0.
	     */
	    if (playerslist[i].id > 0) {
		if (!PlayerIsActive(playerslist[i].id)) {
		    continue;
		}

		/*
		 * Count consecutive turns.
		 */
		if ((playerslist[i].planetsum >= planetsforvictory) && (lastturnnumber < currentturnnumber)) {
		    playerslist[i].consecutiveturns++;
		}
		if (playerslist[i].consecutiveturns > 0) {
		    if (playerslist[i].consecutiveturns >= turnsforvictory) {
			Info("%2u  %-20s    %3u            %2u  WINNER!",
			     playerslist[i].id,
			     RaceShortName(playerslist[i].id, 0),
			     playerslist[i].planetsum,
			     playerslist[i].consecutiveturns);
			snprintf(messages[6 + i], MAXMSGLINE,
				 "%2u %-20s  %3u     %2u WON\x0D",
				 playerslist[i].id,
				 RaceShortName(playerslist[i].id, 0),
				 playerslist[i].planetsum,
				 playerslist[i].consecutiveturns);
			strcat(message, messages[6 + i]);
			/*
			 * Race "i" has fulfilled the winning conditions.
			 */
			playerslist[i].winner = True;
			howmanywinners++;
		    } else {
			Info("%2u  %-20s    %3u             %u",
			     playerslist[i].id,
			     RaceShortName(playerslist[i].id, 0),
			     playerslist[i].planetsum,
			     playerslist[i].consecutiveturns);
			snprintf(messages[6 + i], MAXMSGLINE,
				 "%2u %-20s  %3u     %2u\x0D",
				 playerslist[i].id,
				 RaceShortName(playerslist[i].id, 0),
				 playerslist[i].planetsum,
				 playerslist[i].consecutiveturns);
			strcat(message, messages[6 + i]);
		    }
		} else {
		    if (0 == limittoshow) {
			/*
			 * Report the entire
			 * number of controlled
			 * planets.
			 */
			Info("%2u  %-20s    %3u             -",
			     playerslist[i].id,
			     RaceShortName(playerslist[i].id, 0),
			     playerslist[i].planetsum);
			snprintf(messages[6 + i], MAXMSGLINE,
				 "%2u %-20s  %3u      -\x0D",
				 playerslist[i].id,
				 RaceShortName(playerslist[i].id, 0),
				 playerslist[i].planetsum);
			strcat(message, messages[6 + i]);
		    } else if (((playerslist[i].planetsum * 100) / planetsforvictory) >= limittoshow) {
			/*
			 * Report the number of planets
			 * each player controls if a
			 * distinct percentage is reached.
			 */
			Info("%2u  %-20s    %3u             -",
			     playerslist[i].id,
			     RaceShortName(playerslist[i].id, 0),
			     playerslist[i].planetsum);
			snprintf(messages[6 + i], MAXMSGLINE,
				 "%2u %-20s  %3u      -\x0D",
				 playerslist[i].id,
				 RaceShortName(playerslist[i].id, 0),
				 playerslist[i].planetsum);
			strcat(message, messages[6 + i]);
		    } else {
			;
		    }
		}
	    }
	}
	if (True == reportunownedplanets) {
	    sprintf(messages[6+RACE_NR+1], "\x0D");
	    strcat(message, messages[6+RACE_NR+1]);
	}
    } else {
	Info("Status:\n    No race is controlling more\n    than %u%s of the necessary\n    planets until now.",
	     limittoshow, "%");
	sprintf(messages[6], "Status:\x0D");
	strcat(message, messages[6]);
	sprintf(messages[7], "  No race is controlling more\x0D");
	strcat(message, messages[7]);
	sprintf(messages[8], "  than %u%s of the necessary\x0D",
		limittoshow, "%");
	strcat(message, messages[8]);
	sprintf(messages[9], "  planets until now.\x0D");
	strcat(message, messages[9]);
	if (True == reportunownedplanets) {
	    sprintf(messages[10], "\x0D");
	    strcat(message, messages[10]);
	}
    }
    Info("");
    if (True == reportunownedplanets) {
	Info("Unowned planets: %ld\n", planetunownedsum);
	snprintf(planetunownedmessage, MAXMSGLINE, "Unowned planets: %u\x0D", planetunownedsum);
	strcat(message, planetunownedmessage);
    }


    /*
     * Sort the players list by comparing
     * the id of each player.
     */
    qsort(playerslist, count, sizeof(struct player), playerfind_cmp);


    /*
     * Send report to players.
     */
    messagePtr[0] = message;
    for (race = 0; race <= RACE_NR; race++) {
	if (playerslist[race].id > 0) { /* Race 0 is nothing. */
	    WriteAUXHOSTMessage(race, messagePtr[0]);
	    if (playerslist[race].consecutiveturns >= turnsforvictory) {
		/*
		 * Beep if there is a winner.
		 */
		puts("\a");
	    }
	}
    }


    /*
     * Send message(s) to winner(s).
     */
    if (howmanywinners >= 1) {
	for (race = 0; race <= RACE_NR; race++) {
	    if (playerslist[race].id > 0) { /* Race 0 is nothing. */
		if (True == playerslist[race].winner) {
		    WriteAUXHOSTMessage(playerslist[race].id,
					"(-a0000)<<< Checkvictory - Hurray! >>>\x0D\x0D\x0D    Congratulations!\x0D \x0D    You've fulfilled the\x0D    victory conditions.\x0D");
		    Info("Player %u has fulfilled the victory conditions.", playerslist[race].id);
		}
	    }
	}
    }
    Info("");


    /*
     * Write data file.
     */
    statusfile = OpenOutputFile("checkvic.hst", GAME_DIR_ONLY);
    if (statusfile) {
	if (1 != (fwrite(&currentturnnumber, sizeof(Uns16), 1, statusfile))) {
	    fprintf(errorlogfile ? errorlogfile : stderr, "checkvic.hst: Cannot write current turn number!\n");
	    fclose(statusfile);
	    statusfile = NULL;
	    returncode = EXIT_FAILURE;
	    goto error;
	}
	if (1 != (fwrite(playerslist, sizeof(playerslist), 1, statusfile))) {
	    fprintf(errorlogfile ? errorlogfile : stderr, "checkvic.hst: Cannot write current status!\n");
	    fclose(statusfile);
	    statusfile = NULL;
	    returncode = EXIT_FAILURE;
	    goto error;
	}
	fclose(statusfile);
	statusfile = NULL;
    } else {
	fprintf(errorlogfile ? errorlogfile : stderr,
		"Warning: Cannot open checkvic.hst for writing!\n");
	returncode = EXIT_FAILURE;
	goto error;
    }


    /*
     * Close down.
     */
error:

    if (gLogFile) {
	fclose(gLogFile);
    }
    if (errorlogfile) {
	if (returncode != EXIT_SUCCESS) {
	    fprintf(stderr, "ERROR! Please read the error log file %s'.\n", errorlog);
	}
	fclose(errorlogfile);
	MemFree(errorlog);
    }
    if (message) {
	for (i = 0; i < 20; i++) {
	    MemFree(messages[i]);
	}
	MemFree(message);
    }
    if (statusfile) {
	fclose(statusfile);
    }
    if (True == phostlibstatus) {
	FreePHOSTLib();
    }

    exit(returncode);
}
